# coding: utf-8
# pylint: disable=no-member

u"""Модуль содержит код клиента для запросов в ФМС."""

import datetime
from uuid import uuid4

from django.utils import timezone
from spyne_smev.client import Client
from six import text_type
from suds import WebFault

from .validation import (
    BaseFmsValidator,
    BookRequestValidator,
    ConfigValidator,
    GetResultValidator
)


class FmsRequestConfig(object):
    u"""Класс для хранения настроек взаимодействия."""

    DIGEST_METHOD = "sha1"
    # Тип сообщения (Взаимодействие в рамках оказания государственных услуг)
    TYPE_CODE = u'GSRV'
    # код 2 (Межведомственное взаимодействие)
    EXCHANGE_TYPE = u'2'
    # Тип сообщения
    MESSAGE_STATUS = u'REQUEST'
    VERSION_CODE = '001'
    # Варианты проверки
    REQUEST_TYPE_LIVE_PLACE_CHECK = 1
    REQUEST_TYPE_REAL_PLACE_CHECK = 2

    def __init__(self, wsdl_url,
                 certificate_path, private_key_path, private_key_pass,
                 smev_sender_code, smev_sender_name,
                 recipient_code, recipient_name, system_okato, region_code,
                 service_code, user_id, user_iname, user_oname, user_fname,
                 proxy_params=None):

        u"""
        :param str wsdl_url: URL описания сервиса ФМС (WSDL).
        :param str certificate_path: полный путь до файла сертификата.
        :param str private_key_path: полный путь до файла закрытого ключа.
        :param str private_key_pass: пароль закрытого ключа.
        :param str smev_sender_code: Код системы-инициатора взаимодействия.
        :param str smev_sender_name: Имя системы-инициатора взаимодействия.
        :param str recipient_code: Код системы-получателя сообщения.
        :param str recipient_name: Имя системы-получателя сообщения.
        :param str system_okato: ОКАТО региона установки системы.
        :param str region_code: код региона установки системы.
        :param str service_code: Код государственной услуги.
        :param int user_id: ID пользователя, отправляющего запрос.
        :param unicode user_iname: Имя пользователя, отправляющего запрос.
        :param unicode user_oname: Отчество пользователя.
        :param unicode user_fname: Фамилия пользователя.
        :param dict proxy_params: необязательные параметры прокси-сервера.
        """

        self.wsdl_url = wsdl_url
        self.private_key_path = private_key_path
        self.private_key_pass = private_key_pass
        self.certificate_path = certificate_path
        self.smev_sender_code = smev_sender_code
        self.smev_sender_name = smev_sender_name
        self.recipient_code = recipient_code
        self.recipient_name = recipient_name
        self.system_okato = system_okato
        self.region_code = region_code
        self.service_code = service_code
        self.user_id = user_id
        self.user_iname = user_iname
        self.user_oname = user_oname
        self.user_fname = user_fname

        proxy = {}
        if proxy_params:

            proxy_type = proxy_params['PROXY_TYPE']
            proxy = {
                proxy_type: '%s://%s:%s@%s:%s/' % (
                    proxy_type,
                    proxy_params['PROXY_USER'], proxy_params['PROXY_PWD'],
                    proxy_params['PROXY_HOST'], proxy_params['PROXY_PORT']
                )
            }

        self.proxy = proxy


class FmsClient(object):
    u"""Клиент для запросов в ФМС."""

    # коды возврата веб-сервиса
    RC_REQUEST_ACCEPTED = 0  # Заявка принята
    RC_ORGANIZATION_NOT_FOUND = 1  # Организация не найдена
    RC_REQUEST_UNAVAILABLE = 3  # Данный тип запроса недоступен для организации
    RC_FAST_REPLY_UNAVAILABLE = 4  # Быстрый ответ невозможен
    RC_FLK_FAILED = 5  # Ошибка ФЛК параметров запроса
    RC_TASK_NOT_FOUND = 6  # Задача не найдена
    RC_WRONG_REGION = 7  # Регион в составе запроса не соотв. справочному
    RC_REQUEST_DUPLICATE = 8  # Дублирующий запрос
    RC_WRONG_SIGNATURE = 9  # Ошибка проверки ЭЦП
    RC_HEADER_NOT_FOUND = 10  # Отсутствует секция заголовка СМЭВ
    RC_REQUEST_INACTIVE = 11  # Заказанная версия типа запроса не активна
    RC_UNEXPECTED_ERROR = -1  # Непредвиденная ошибка

    config = None
    smev_client = None

    def __init__(self, config):
        u"""Принимает и сохраняет параметры конфигурации.

        :param FmsRequestConfig config: Экземпляр конфигурации запроса.
        """

        assert isinstance(config, FmsRequestConfig), type(config)

        self.last_sent = None
        self.last_received = None
        self.config = config

        config_validator = ConfigValidator()
        config_validator.validate(config)

        # Внутренний клиент для взаимодействия со СМЭВ.
        self.smev_client = Client(
            url=self.config.wsdl_url,
            private_key_path=self.config.private_key_path,
            private_key_pass=self.config.private_key_pass,
            certificate_path=self.config.certificate_path,
            digest_method=self.config.DIGEST_METHOD,
            proxy=self.config.proxy
        )

    class Request(object):
        u"""Базовый класс запросов."""

        def __init__(self, request_type):
            u"""
            :param str request_type: Требуемый запрос (bookRequest, getResult).
            """
            self.config = None
            self.params = None
            self.validator = BaseFmsValidator()
            self.request_type = request_type

        def configure(self, config):
            u"""Устанавливает настройки отправки запроса (url, cert...).

            :param FmsRequestConfig config: Экземпляр конфигурации запроса.
            """
            self.config = config

        def _set_tech_params(self, msg):
            u"""Метод записывает в сообщение служебную информацию.

            :param msg: сообщение в формате suds.
            """

            # Код/Название системы-инициатора взаимодействия (потребителя)
            msg.Message.Sender.Code = self.config.smev_sender_code
            msg.Message.Sender.Name = self.config.smev_sender_name

            # Код/Название системы-получателя сообщения (поставщика)
            msg.Message.Recipient.Code = self.config.recipient_code
            msg.Message.Recipient.Name = self.config.recipient_name

            # Код/Название системы, инициировавшей цепочку из нескольких
            # запросов-ответов, объединенных единым процессом
            msg.Message.Originator.Code = self.config.smev_sender_code
            msg.Message.Originator.Name = self.config.smev_sender_name

            msg.Message.TypeCode = self.config.TYPE_CODE

            # Дата и время формирования запроса
            msg.Message.Date = datetime.datetime.now(
                timezone.get_current_timezone()
            ).isoformat()

            msg.Message.ExchangeType = self.config.EXCHANGE_TYPE
            msg.Message.ServiceCode = self.config.service_code
            msg.Message.Status = self.config.MESSAGE_STATUS
            msg.Message.CaseNumber = str(uuid4())

            app_data = msg.MessageData.AppData

            # код из префикса R + 4 первых цифры ОКАТО региона + 6 нулей
            app_data.user.organization = u'R{0}000000'.format(
                self.config.system_okato[0:4]
            )

            # Данные исполнителя. ФИО пользователя, отправившего запрос
            app_data.user.person.id = self.config.user_id
            app_data.user.person.firstName = self.config.user_iname
            app_data.user.person.secondName = self.config.user_oname
            app_data.user.person.lastName = self.config.user_fname

        def _set_request_params(self, msg):
            u"""Метод записывает в сообщение параметры запроса.

            :param msg: сообщение в формате suds.
            """
            raise NotImplementedError

        def execute(self, smev_client, params):
            u"""Выполнение запроса.

            :param smev_client: spyne-smev Client.
            :param dict params: Информация, передаваемая в сообщении.
            """
            self.params = params

            self.validator.validate(self.params)

            msg = smev_client.factory.create(self.request_type)

            self._set_tech_params(msg)
            self._set_request_params(msg)

            # Выполнение требуемого запроса. По-умолчанию запросы запускаются:
            # smev_client.service.bookRequest(msg.Message, msg.MessageData)
            # или smev_client.service.getResult(msg.Message, msg.MessageData)
            service_request = getattr(smev_client.service, self.request_type)
            result = service_request(msg.Message, msg.MessageData)

            last_sent = text_type(smev_client.last_sent())
            last_received = text_type(smev_client.last_received())

            return result, last_sent, last_received

    class BookRequest(Request):
        u"""Запрос bookRequest."""

        def __init__(self, request_type):
            u"""
            :param str request_type: Требуемый запрос (bookRequest, getResult).
            """
            super(FmsClient.BookRequest, self).__init__(request_type)
            self.validator = BookRequestValidator()

        def _set_tech_params(self, msg):
            u"""
            :param msg: сообщение в формате suds.
            """
            super(FmsClient.BookRequest, self)._set_tech_params(msg)

            app_data = msg.MessageData.AppData
            app_data.versionCode = self.params.pop(
                'versionCode', self.config.VERSION_CODE)

        def _set_request_params(self, msg):
            u"""Параметры проверки по адресу.

            :param msg: сообщение в формате suds.

            :param declaration: Заявление
            :type declaration: kinder.core.declaration.models.Declaration
            """

            parameters = msg.MessageData.AppData.parameters
            app_data = msg.MessageData.AppData

            app_data.serviceCode = self.params.pop('serviceCode')

            for param in self.params:
                parameters.append({'name': param, 'value': self.params[param]})

    class GetResult(Request):
        u"""Запрос getResult."""

        def __init__(self, request_type):
            u"""
            :param str request_type: Требуемый запрос (bookRequest, getResult).
            """
            super(FmsClient.GetResult, self).__init__(request_type)
            self.validator = GetResultValidator()

        def _set_request_params(self, msg):
            u"""Заполняет раздел Message.
            :param msg: раздел Body.getResult
            """
            # В соответствии с классификатором "Мнемоники статусов сообщения"
            # из документации СМЭВ должна быть указана мнемоника PING (Запрос)
            msg.Message.Status = u'PING'
            # Идентификатор сообщения-запроса, инициировавшего взаимодействие
            msg.Message.RequestIdRef = self.params['RequestIdRef']
            # Идентификатор сообщения-запроса, инициировавшего цепочку
            msg.Message.OriginRequestIdRef = self.params['OriginRequestIdRef']
            msg.MessageData.AppData.taskId = self.params['taskId']

    def _create_request(self, request_type):
        u"""Возвращает инстанс запроса в зависимости от типа запроса.

        :param str request_type: Требуемый запрос (bookRequest, getResult).
        """
        if request_type == 'bookRequest':
            return self.BookRequest(request_type)
        elif request_type == 'getResult':
            return self.GetResult(request_type)

    def request(self, request_type, params):
        u"""Выполнение запроса к ФМС.
        :param str request_type: Требуемый запрос (bookRequest, getResult).
        :param dict params: Информация, передаваемая в сообщении.
        """
        request = self._create_request(request_type)
        request.configure(self.config)
        try:
            result, last_sent, last_received = request.execute(
                self.smev_client, params
            )
        except WebFault as e:
            self.last_sent = self.smev_client.last_sent()
            self.last_received = e.document
            raise
        else:
            self.last_sent = last_sent
            self.last_received = last_received

        return result
